/*
 *  JDynamiTe - Dynamic Template in Java
 *  Copyright (C) 2001 Christophe Bouleau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package cb.jdynamite.analyser;

import java.io.*;
import gnu.regexp.*;
import cb.jdynamite.*;

/**
 * This class analyses the input template document and builds the JDynamiTe document structure,
 * which is made up of "ITemplateElements".<br>
 * A JDynamiTe document is structurally a tree where the nodes are "DynamicElement" objects,
 * and where the leaves are the other objects.<br>
 * An analyser is attached to an "ITemplateDocument" (JDynamiTe object),
 * which is the "root" Dynamic Element of the tree.
 * 
 * @see cb.jdynamite.JDynamiTe
 */
public class DefaultAnalyser implements ITemplateAnalyser {
    private ITemplateDocument rootDoc;
    private BufferedReader input;
    private RE varRegExp;
    private RE dynBeginRegExp;
    private RE dynEndRegExp;
    private RE ignBeginRegExp;
    private RE ignEndRegExp;
    private boolean debug = false;
    
    /**
      HTML default "Variable" tag.
      This is the regular expression for a word (within brackets) made up of alplanumeric (and '.', '_', ':', '-') character(s).
      Examples : "Table2.name" or "page-title".<br>
      "[\w._:-]+" is the regular expression for a word made up of alplanumeric (and '.', '_', ':', '-') character(s).
      <br>Important: the variable name (the key) is this word wich matches the first "regexp" subexpression.
      That's why this word should be enclosed within brackets.
      <br>See GNU regexp package documentation for more details on regular expressions.
    */
    public static String HTML_VARIABLE_TAG = "\\{([\\w._:-]+)\\}";
    /**
      HTML default "Begin Dynamic Element" tag.
      This regular expression is : <pre>
      "&lt;!-- *BEGIN *DYNAMIC.*: *([\w._:-]+) *--&gt;".
      </pre>For example:<pre>
      "&lt;-- BEGIN DYNAMIC : myList --&gt;".
      </pre>or<pre>
      "&lt;-- BEGIN DYNAMIC BLOCK: myList --&gt;".
      </pre>Important: the element name is this word wich matches the first "regexp" subexpression.
      That's why this word should be enclosed within brackets.
      <br>See GNU regexp package documentation.
    */
    public static String HTML_BEGIN_DYNAMIC_TAG = "<!-- *BEGIN *DYNAMIC.*: *([\\w._:-]+) *-->";
    /**
      HTML default "End Dynamic Element" tag.
      This regular expression is : <pre>
      "&lt;-- *END *DYNAMIC.*--&gt;".
      </pre>For example:<pre>
      "&lt;-- END DYNAMIC : myList --&gt;".
      </pre>or<pre>
      "&lt;-- END DYNAMIC BLOCK: myList --&gt;".
      </pre>See GNU regexp package documentation.
    */
    public static String HTML_END_DYNAMIC_TAG = "<!-- *END *DYNAMIC.*-->";
    /**
     * HTML default "Begin Ignored Element" tag.
     * This regular expression is : <pre>
     * "&lt;!-- *BEGIN *IGNORED.*: *([\w._:-]+) *--&gt;".
     * </pre>For example:<pre>
     * "&lt;-- BEGIN IGNORED : myLinesExample --&gt;".
     * </pre>See GNU regexp package documentation.
     * 
     * @since JDynamiTe 1.1
     */
    public static String HTML_BEGIN_IGNORED_TAG = "<!-- *BEGIN *IGNORED.*: *([\\w._:-]+) *-->";
    /**
     * HTML default "End Ignored Element" tag.
     * This regular expression is : <pre>
     * "&lt;-- *END *IGNORED.*--&gt;".
     * </pre>For example:<pre>
     * "&lt;-- END IGNORED : myList --&gt;".
     * </pre>See GNU regexp package documentation.
     * 
     * @since JDynamiTe 1.1
     */
    public static String HTML_END_IGNORED_TAG = "<!-- *END *IGNORED.*-->";
    
    public static String XML_VARIABLE_TAG = "\\{([\\w._:-]+)\\}";
    public static String XML_BEGIN_DYNAMIC_TAG = "<!-- *BEGIN *DYNAMIC.*: *([\\w._:-]+) *-->";
    public static String XML_END_DYNAMIC_TAG = "<!-- *END *DYNAMIC.*-->";
    public static String SCRIPT_VARIABLE_TAG = "\\$([\\w._:-]+)";
    public static String SCRIPT_BEGIN_DYNAMIC_TAG = "#+ *BEGIN *DYNAMIC.*: *([\\w._:-]+) *";
    public static String SCRIPT_END_DYNAMIC_TAG = "#+ *END *DYNAMIC.*";

    public void DefaultAnalyser() {
    }

    public void analyse(ITemplateDocument rootDocument, IDynamicElement rootElem, BufferedReader inputText) throws IOException {
        rootDoc = rootDocument;
        input = inputText;
        initRegExp();
        doAnalyse(rootElem, 0);
    }

    public RE getVariableRegExp() {
        return varRegExp;
    }
    public void setVariableRegExp(String variableRegExp) throws REException {
        varRegExp = new RE(variableRegExp);
    }

    public RE getDynamicBeginRegExp() {
        return dynBeginRegExp;
    }
    public void setDynamicBeginRegExp(String dynamicBeginRegExp) throws REException {
        dynBeginRegExp = new RE(dynamicBeginRegExp);
    }

    public RE getDynamicEndRegExp() {
        return dynEndRegExp;
    }
    public void setDynamicEndRegExp(String dynamicEndRegExp) throws REException {
        dynEndRegExp = new RE(dynamicEndRegExp);
    }

    public RE getIgnoredBeginRegExp() {
        return ignBeginRegExp;
    }
    public void setIgnoredBeginRegExp(String ignoredBeginRegExp) throws REException {
        ignBeginRegExp = new RE(ignoredBeginRegExp);
    }

    public RE getIgnoredEndRegExp() {
        return ignEndRegExp;
    }
    public void setIgnoredEndRegExp(String ignoredEndRegExp) throws REException {
        ignEndRegExp = new RE(ignoredEndRegExp);
    }

    public boolean getDebug() {
        return debug;
    }
    public void setDebug(boolean debugMode) {
        debug = debugMode;
    }

    private void initRegExp() {
        varRegExp = null;
        dynBeginRegExp = null;
        dynEndRegExp = null;
        ignBeginRegExp = null;
        ignEndRegExp = null;
        try {
            setVariableRegExp(DefaultAnalyser.HTML_VARIABLE_TAG);
            setDynamicBeginRegExp(DefaultAnalyser.HTML_BEGIN_DYNAMIC_TAG);
            setDynamicEndRegExp(DefaultAnalyser.HTML_END_DYNAMIC_TAG);
            setIgnoredBeginRegExp(DefaultAnalyser.HTML_BEGIN_IGNORED_TAG);
            setIgnoredEndRegExp(DefaultAnalyser.HTML_END_IGNORED_TAG);
        }
        catch (REException e) {
            e.printStackTrace();
        }
    }

    private void addMixed(IDynamicElement dynElem, StringBuffer content) {
        REMatchEnumeration matchEnum = varRegExp.getMatchEnumeration(content);
        int constantStart = 0;
        int endMatch;
        while (matchEnum.hasMoreElements()) {
            REMatch match = matchEnum.nextMatch();
            if (constantStart < match.getStartIndex()) {
                ConstantElement constElem= new ConstantElement(content.substring(constantStart,
                                                                                 match.getStartIndex()));
                dynElem.addElement(constElem);
            }
            constantStart = match.getEndIndex();
            // Important: the variable name (the key) is this word wich matches the first
            // "regexp" subexpression. That's why this word should be enclosed within brackets in varRegExp.
            String key = match.toString(1);
            // Equivalent : match.substituteInto("$1");
            if (key == null) {
                System.err.println("Can not get Variable element name from this string: " + match.toString());
                key = "___Undefined Name___";
            }
            VariableElement varElem = new VariableElement(key);
            dynElem.addElement(varElem);
            // Add this variable to the root document hashtable
            rootDoc.setVariable(key, "");
        }
        if (constantStart < content.length()) {
            ConstantElement constElem= new ConstantElement(content.substring(constantStart));
            dynElem.addElement(constElem);
        }
        
        content.delete(0, content.length());
        /***
        if (content.length() != 0) {
            MixedElement mixedElem = new MixedElement(content.toString());
            dynElem.addElement(mixedElem);
        }
        ***/
    }

/*
    private String getDynamicElementNameFromString(String line) {
        int startIndex = line.lastIndexOf(':');
        if (startIndex != -1) {
            int lastIndex = line.lastIndexOf("-->");
            if (lastIndex != -1) {
                return (line.substring(startIndex + 1, lastIndex)).trim();
            }
        }
        System.err.println("Can not get element name from this line: " + line);
        return "___Undefined Name___";   
    }
*/

    protected void doAnalyse(IDynamicElement dynElem, int depth) {
        StringBuffer mixed = new StringBuffer();
        boolean inIgnoredElement = false;
        boolean theEnd = false;
        while (!theEnd) {
            String line = null;
            try {
                line = input.readLine();
            }
            catch (IOException e) {
                System.err.println(e.getMessage());
                break;
            }
            if (line == null) {
                addMixed(dynElem, mixed);
                break;
            }
            REMatch match;
            if (inIgnoredElement) {
                match = ignEndRegExp.getMatch(line);
                if (match != null) {
                    // end of an ignored element
                    inIgnoredElement = false;
                    if (debug) {
                        System.err.println("<<End Ignore\n");
                    }
                }
                else if (debug) {
                    System.err.println("ignore[" + line + "]\n");
                }
                continue;
            }
            match = dynBeginRegExp.getMatch(line);
            if (match != null) {
                // begin of a dynamic element.
                // first add the current mixed element
                addMixed(dynElem, mixed);
                // then build a new dynamic element
                String name = match.toString(1); // first subexpression
                if (name == null) {
                    System.err.println("Can not get dynamic element name from this line: " + line);
                    name = "___Undefined Name___";
                }
                DefaultDynamicElement newDyn = new DefaultDynamicElement(name);
                dynElem.addElement(newDyn);
                rootDoc.recordDynElem(name, newDyn); // record it in ITemplateDocument root document
                doAnalyse(newDyn, depth + 1); // recursively analyse this new bloc
            }
            else {
                match = dynEndRegExp.getMatch(line);
                if (match != null) {
                    // end of the dynamic element
                    addMixed(dynElem, mixed);
                    theEnd = true;
                }
                else {
                    match = ignBeginRegExp.getMatch(line);
                    if (match != null) {
                        // begin of an ignored element
                        inIgnoredElement = true;
                        if (debug) {
                           System.err.println("Begin Ignore>>\n");
                        }
                    }
                    else {
                        // just a line in the dynamic element
                        mixed.append(line + '\n');
                    }
                }
            }
        }
    }
}
